/* File I/O for Node */
/* CPS interface */

var fs=require('fs')
  , C=require('./continuable')
  , E=require('./either')
  , either=E.either, Right=E.Right, Left=E.Left

/* Wrappers returning Either types for node's 'fs' module */

exports.rename=rename
exports.truncate=truncate
exports.chmod=chmod
exports.stat=stat
exports.unlink=unlink
exports.rmdir=rmdir
exports.mkdir=mkdir
exports.readdir=readdir
exports.close=close
exports.open=open
exports.write=write
exports.read=read

/* Exports */

// readFile :: (path::String, encoding::String) → Continuable Either Error String
// writeFile :: (path::String, data::String, encoding::Opt String, mode::Opt Int) → Continuable Either Error _
// appendFile :: (String, String, Opt String, Opt Int) → Continuable Either Error _
// touch :: (String, mode::Opt Int) → Continuable Either Error _
// copyFile :: (src::String, dest::String) → Continuable Either Error _
// streamFile :: (path::String, encoding::String) → Stream
// streamFileBy :: (path::String, encoding::String, separator::String) → Stream
// streamLines(path,encoding) equivalent to streamFileBy(path,encoding,"\n")
// writeStream :: (path::String, encoding::String, mode::Int) → Stream → Continuable Either Error _
// appendStream :: (path::String, encoding::String, mode::Int) → Stream → Continuable Either Error _
// bufferedStream :: Stream → Stream
exports.readFile=readFile
exports.writeFile=writeFile
exports.appendFile=appendFile
exports.touch=touch

exports.copyFile=copyFile

exports.streamFile=streamFile
exports.streamFileBy=streamFileBy
exports.streamLines=streamLines
exports.writeStream=writeStream
exports.appendStream=appendStream
exports.bufferedStream=bufferedStream

// type Stream = ( Consumer StreamMsg | StreamCtrlMsg ) → _
// type StreamMsg = "error" | "chunk", String, Int | "eof"
// type StreamCtrlMsg = "next" | "close"
// type Consumer a = a → _
// type Opt a :: A | Undefined

// Consumers and Continuations are called only for their side-effects, hence the return type "_", meaning "unused".
// A consumer is called repeatedly to consume a stream of events, while a continuation is called only once.

/* Helper functions for Either */

function Either(left,right){return left?[0,left]:[1,right]}

exports.bracket=bracket
// bracket :: (Continuable Either Error a, a → Continuable Either Error b, a → Continuable _) → Continuable Either Error b
function bracket(before,during,after){return function(cont){
  before(E.replaceRight(cont)
                       (function(a){during(a)(function(either){after(a)(function(){cont(either)})})}))}}

// withFd :: (path::String, flags::Int, mode::Int) → (Fd → Continuable a) → Continuable a
exports.withFd=withFd
function withFd(path,flags,mode){return function(fd_cont){return function(cont){
  open(path,flags,mode)(function(fd){
    fd_cont(fd)(function(a){finish(a)})})
  function finish(a){
    close(fd)(function(){cont(a)})}}}}

/* Wrappers returning Either types for node's 'fs' module */

// rename :: (String, String) → Continuable Either Error _
function rename(from,to){return function(cont){
  fs.rename(from,to,function(err){cont(err?Left(err):Right())})}}

// truncate :: (Fd, length::Int) → Continuable Either Error _
function truncate(fd,len){return function(cont){
  fs.truncate(fd,len,function(err){cont(err?Left(err):Right())})}}

// chmod :: (path::String, mode::Int) → Continuable Either Error _
function chmod(path,mode){return function(cont){
  fs.chmod(path,mode,function(err){cont(err?Left(err):Right())})}}

// stat :: String → Continuable Either Error StatResult
function stat(path){return function(cont){
  fs.stat(path,function(err,stat){cont(Either(err,stat))})}}

// unlink :: String → Continuable Either Error _
function unlink(path){return function(cont){
  fs.unlink(path,function(err){cont(err?Left(err):Right())})}}

// rmdir :: String → Continuable Either Error _
function rmdir(path){return function(cont){
  fs.rmdir(path,function(err){cont(err?Left(err):Right())})}}

// mkdir :: (String, Int) → Continuable Either Error _
function mkdir(path,mode){return function(cont){
  fs.mkdir(path,mode,function(err){cont(err?Left(err):Right())})}}

// readdir :: String → Continuable Either Error [String]
function readdir(path){return function(cont){
  fs.readdir(path,function(err,files){cont(Either(err,files))})}}

// close :: Fd → Continuable Either Error _
function close(fd){return function(cont){
  fs.close(fd,function(err){cont(err?Left(err):Right())})}}

// open :: (String, Int, Int) → Continuable Either Error Fd
function open(path,flags,mode){return function(cont){
  fs.open(path,flags,mode,function(err,fd){cont(Either(err,fd))})}}

// write :: (Fd, String, pos::Int, encoding::String) → Continuable Either Error Int
function write(fd,data,pos,encoding){return function(cont){
  fs.write(fd,data,pos,encoding,function(err,bytes){cont(Either(err,bytes))})}}

// read :: (Fd, length::Int, position::Int, encoding::String) → Continuable Either Error String
function read(fd,len,pos,encoding){return function(cont){
  fs.read(fd,len,pos,encoding,function(err,data,bytes){cont(Either(err,[data,bytes]))})}}

// closeThrowing :: Fd → Continuable Right _
// throws an exception if the close(2) call returns an error
// useful when the successful return is uninteresting and failure indicates a major problem
function closeThrowing(fd){return function(cont){fs.close(fd,function(x){if(x)throw x;cont([1])})}}

// writeFile takes a path, a string, an encoding, and a mode.
// It will create or replace the contents of an existing file.
// If the file is created, the permissions are taken from `mode` (and the process umask).
// It succeeds when the data has been written to the file (as reported by the OS) and the file handle has been closed.

//var writeFlags=process.O_CREAT|process.O_WRONLY|process.O_TRUNC
var writeFlags="w"
  , defaultMode=0x1B6 // 0666

// writeFile :: (path::String, data::String, encoding::Opt String, mode::Opt Int) → Continuable Either Error _
function writeFile(path,data,encoding,mode){encoding=encoding||'utf-8';mode=mode||defaultMode
  return bracket(open(path,writeFlags,mode)
                ,writeChunks(data,encoding)
                ,closeThrowing)}

//var appendFlags=process.O_CREAT|process.O_APPEND|process.O_WRONLY
var appendFlags="a"

// appendFile :: (String, String, Opt String, Opt Int) → Continuable Either Error _
function appendFile(path,data,encoding,mode){encoding=encoding||'utf-8';mode=mode||defaultMode
  return bracket(open(path,appendFlags,mode)
                ,writeChunks(data,encoding)
                ,closeThrowing)}

// touch :: (String, Opt Int) → Continuable Either Error _
function touch(path,mode){return writeFile(path,"","ascii",mode)}

// writeChunks :: String, String → Fd → Continuable Either Error _
function writeChunks(data,encoding){return function(fd){return function loop(cont){
  if(encoding!='binary' && encoding!='ascii')throw new Error("unsupported encoding "+encoding) // XXX see issue in writeStream below
  write(fd,data,null,encoding)(E.replaceRight(cont)(function(nbytes){
    if(nbytes<data.length){
      data=data.slice(nbytes)
      loop(cont)}
    else cont([1])}))}}}

// A stream is a stream of chunks, which can either be read through to EOF or closed at any time.
// A stream is consumed by passing a consumer to it, which is a function which will be called with stream messages.
// A stream message is either a 'chunk', 'eof', or 'error' message.
// After each 'chunk' message, the consumer must either send a 'next' or 'close' message to the stream.
// If EOF is reached, or an error is generated, the consumer need not send a close message, as the stream will close itself before sending these events.

// readFile :: (path::String, encoding::String) → Continuable Either Error String
function readFile(path,encoding){return function(cont){var s,buf=[]
  s=streamFile(path,encoding)
  s(function(m,x,y){switch(m){
    case 'error':cont([0,x]);break
    case 'chunk':buf.push(x);s('next');break
    case 'eof':cont([1,buf.join('')])}})}}

var readFlags="r"

// streamFile :: String, String → Stream
// we now queue reads, in case there is already a pending read when a 'next' message comes in
// however, perhaps we should just return an error instead to help catch programming errors
function streamFile(path,encoding){var consumer,fd,blksize,closed,pending=0,pos=0;return function(m){
  if(!consumer){ // this is the first call to the stream, m is the consumer
    if(typeof m != 'function')throw new Error('stream called with non-function as first argument')
    consumer=m
    stat(path)(either(err,function(stat_result){
      blksize=stat_result.blksize||4096
      open(path,readFlags,0666)(either(err,function(fd_){fd=fd_;pending++;readAndEmit(fd)}))}))}
  else{switch(m){
    case 'next':if(closed){consumer('error',new Error('sent "next" message to a closed stream!'))}
                else{pending++;if(pending==1){readAndEmit(fd)}}break
    case 'close':close()}}
  function err(e){consumer('error',e)}
  function close(){closeThrowing(fd)(function(){closed=true})}
  function readAndEmit(fd){
    read(fd,blksize,pos,encoding)(either(function(e){closeThrowing(fd)(function(){consumer('error',e)})}
                                        ,function(a){if(a[1]){consumer('chunk',a[0],a[1]);pos+=a[1]}
                                                     else    {consumer('eof');close()}
                                                     pending--
                                                     if(pending)readAndEmit(fd)}))}}}

// streamLines :: (path::String, encoding::String) → Stream
function streamLines(path,encoding){return streamFileBy(path,encoding,'\n')}

// streamFileBy :: (path::String, encoding::String, separator::String) → Stream
// like streamFile, but instead of arbitrary sized chunks, each chunk (except for the last) is terminated by `separator`
// a possible optimization here would be to use an array as a buffer instead of appending to a string
function streamFileBy(path,encoding,sep){var l=sep.length,consumer,chunkstream,buf="",closed=false,eof=false
  return function(m){
    if(!consumer){
      if(typeof m != 'function')throw new Error('first call to stream must provide a consumer')
      consumer=m
      chunkStream=streamFile(path,encoding)
      chunkStream(function(m,x,y){switch(m){
        case 'error':consumer('error',x);break
        case 'chunk':buf+=x;emitOrKeepReading();break
        case 'eof':eof=true;consumer('chunk',buf)}})}
    else{switch(m){
                  // we catch the read-from-closed case early here to alert the user to a programming error
                  // otherwise it would still be possible to read any remaining chunks from the buffer we already have
      case 'next':emitOrKeepReading();break
      case 'close':closed=true;chunkStream('close')}}}
  function emitOrKeepReading(){var pos,chunk
    if(closed){consumer('error',new Error('sent "next" message to a closed stream!'));return}
    if(eof){closed=true;consumer('eof');return}
    pos=buf.indexOf(sep)
    if(pos==-1){chunkStream('next')} // keep reading
    else{ // emit
      chunk=buf.substring(0,pos+l)
      buf=buf.substring(pos+l)
      consumer('chunk',chunk,pos+l)}}}

// writeStream :: (path::String, encoding::String, mode::Int) → Stream → Continuable Either Error _
function writeStream(path,encoding,mode){mode=mode||defaultMode;return function(stream){
  // see http://nodejs.debuggable.com/2010-01-04.txt starting at "[18:49]"
  if(encoding!='binary' && encoding!='ascii')throw new Error("unsupported encoding "+encoding)
  return bracket(open(path,writeFlags,mode)
                ,streamToFd(stream,encoding)
                ,closeThrowing)}}

// helper function for writeStream and appendStream
// streamToFd :: (Stream, encoding::String) → Int → Continuable Either Error _
function streamToFd(stream,encoding){return function(fd){return function(cont){var writing=false
  stream(function(m,x,y){switch(m){
    case 'error':cont([0,x]);break
    case 'chunk':if(writing)throw new Error('streamToFd: stream did not wait for \'next\' message before sending next chunk!')
                 writing=true;writeChunks(x,encoding)(fd)(E.replaceRight(cont)(function(){writing=false;stream('next')}));break
    case 'eof':cont([1])}})}}}

// appendStream :: (path::String, encoding::String, mode::Int) → Stream → Continuable Either Error _
function appendStream(path,encoding,mode){mode=mode||defaultMode;return function(stream){
  if(encoding!='binary' && encoding!='ascii')throw new Error("unsupported encoding "+encoding)
  return bracket(open(path,appendFlags,mode)
                ,streamToFd(stream,encoding)
                ,closeThrowing)}}

// copyFile :: (src::String, dest::String) → Continuable Either Error _
// uses streams so as to copy files of arbitrary size in constant memory
function copyFile(src,dest){
  return writeStream(dest,"binary")(streamFile(src,"binary"))}

function assert(b,m){if(!b)throw new Error('assertion failed: '+m)}

// bufferedStream :: Stream → Stream
// bufferedStream buffers an input stream by requesting the next chunk synchronously each time a chunk is received.
// This allows the input stream to emit chunks at any rate it likes, without concern about overflowing a writing process that is the ultimate consumer, and lifts the requirement to listen for 'next' messages which is otherwise required of well-behaved streams.
// Chunks will be requested from the input stream (i.e. it will receive 'next' messages) synchronously, and the input stream may send chunks at any rate it likes, synchronously or asynchronously.
// Chunks will be buffered as necessary and then streamed out to the final consumer as they are requested.
// Close messages from the final consumer will be propagated back to the input stream.
// Chunk, eof, and error messages from the input stream will be buffered and propagated forward in the order they are received.
// Chunks received and buffered before an error message will still be sent to the consumer.
function bufferedStream(sIn){var consumer,queue,consumerIsWaiting,inputIsClosed
  queue=[]
  return function(x){
    if(!consumer){
      if(typeof x != 'function')throw new Error('bufferStream: called without a consumer function')
      consumer=x
      consumerIsWaiting=true // the consumer "is waiting" before the first chunk is sent, and subsequently when it has sent a 'next' to which we have not yet replied
      inputIsClosed=false
      sIn(function(m,x,y){switch(m){
        case 'chunk':
          if(inputIsClosed)throw new Error('A closed stream cannot continue to send messages')
          queue.push(m,x)
          if(consumerIsWaiting)send()
          sIn('next');break
        case 'eof':
          inputIsClosed=true
          queue.push(m)
          if(consumerIsWaiting)send()
          break
        case 'error':
          inputIsClosed=true
          queue.push(m,x)
          if(consumerIsWaiting)send()}})}
    switch(x){
      case 'next':
        if(queue.length)send()
        else consumerIsWaiting=true
        break
      case 'close':
        sIn('close')
        inputIsClosed=true}}
  function send(){var m,x
    assert(queue.length,"queue.length")
    m=queue[0]
    switch(m){
      case 'error':
      case 'chunk': assert(queue.length>1,"queue length > 1");x=queue[1];queue=queue.slice(2)
      case 'eof': queue=queue.slice(1)}
    consumerIsWaiting=false
    consumer(m,x)}}
