#!/usr/bin/env node

// the working directory should be the location of httpd.js, for relative references to various files such as the HTTP error template file, and module references in dispatch trees.
process.chdir(__dirname)

var http=require('http')
  , fs=require('fs')
  , sys=require('sys')
  , opts=require('opts'),options
  , file=require('./deps/fs-continuable')
  , E=require('./deps/either'),either=E.either
  , Left=E.Left,Right=E.Right,applyRight=E.__applyRight
  , C=require('./deps/continuable')
  , handlers=require('./handlers')
  , rfc=require('./rfc_util')
  , nhttpd={}
  , dispatchTree // the server's tree for dispatching requests to handlers
  , treeFallthrough // fallthrough handlers per tree type
  , server // the http.Server object
  , port=8080
  , domain=''
    // dispatch map types **NB** if updated, also update typeNames below
  , HANDLER=0    // A request handler (terminates descent and has no children)
  , METHOD=1     // branch on the request method (GET, HEAD, ...)
  , PATHHEAD=2   // branch on the next path segment
  , UA=3         // branch on User-Agent string
  , AUTHN=4      // branch on HTTP Authentication
  , COOKIE=5     // branch on a particular cookie value
  , REMOTE=6     // branch on the remote IP address
  , FILENAME=7   // branch on the last path segment (the "filename", e.g. "baz" in "foo/bar/baz")
  , QUERY=8      // branch on the query string (everything in the path after the first '?')
  , HOST=9       // branch on the Host header ("name-based virtual hosting")
    // function to distinguish between HANDLER and the branch types
  , isBranchType=function(tree){return tree[0]>HANDLER}
    // dispatch tree type names
    // the dispatch tree type numeric constants above are for internal use 
    // only, for the external configuration API we use friendly string names 
    // rather than numeric constants.
  , typeNames=['RequestHandler'
              ,'MethodBranch'
              ,'PathBranch'
              ,'UserAgentBranch'
              ,'AuthBranch'
              ,'CookieBranch'
              ,'RemoteAddressBranch'
              ,'FilenameBranch'
              ,'QueryBranch'
              ,'HostBranch']
  , debug=1 // debug level 0-3

// for debugging
GLOBAL.dump=dump
function dump(x){process.binding('stdio').writeError(x===undefined?'undefined\n':sys.inspect(x,null,1)+'\n')}

// for logging, at log level 1, 2, or 3 (increasing verbosity)
// 1 should be used for most important messages and all failures, 2 and 3 for increasing levels of debugging output
GLOBAL.dbg1=debug>0?dump:function(){}
GLOBAL.dbg2=debug>1?dump:function(){}
GLOBAL.dbg3=debug>2?dump:function(){}

// TODO: stick this somewhere else
GLOBAL.assert=assert
function assert(b,s){if(!b)throw new Error('assertion failed: '+s)}

options=
[{short:'p'
 ,long:'port'
 ,description:'port to listen for HTTP requests on'
 ,value:true
 }
]

opts.parse(options)

/* Determine the port to listen on */

// By default listen on port 80, but can be overridden by "-p PORT" on the command line.
port=opts.get('port')||port

if(!port) throw new Error('-p option requires a PORT argument')

/* Default Dispatch Map */

// The initial dispatch tree only maps /conf/dispatch, and only from 
// localhost.
// To non-localhost requests, anything below /conf will be 403'd

// Once nhttpd is running, you must connect to it from localhost to 
// set up the desired dispatch tree to serve other requests.

dispatchTree=ChoicePoint(PATHHEAD)
appendPattern(dispatchTree,
  {test:'conf/'
  ,then:[REMOTE,[{test:'127.0.0.1'
                 ,then:[PATHHEAD,[{test:'dispatch/'
                                  ,then:[PATHHEAD,[{generator:dispatchTreeDispatchGenerator(dispatchTree)
                                                   ,uri:'/builtins/generators/dispatchTree'}]]}]]}]]})

// Some constructors as a alternative to direct construction of objects.
// This feature might go away.

nhttpd.ChoicePoint=ChoicePoint
nhttpd.Handler=Handler
nhttpd.Pattern=Pattern
nhttpd.appendPattern=appendPattern
nhttpd.typeNames

function ChoicePoint(type,cn){
 if(typeof type=='string')type=typeNames.indexOf(type)
 return [type,cn||[]]}

function Handler(f){
 return [HANDLER,f]}

function Pattern(test,consequent){
 return {test:test,then:consequent}}

function appendPattern(choicePoint,node){
 choicePoint[1].push(node)
 return choicePoint}

// dispatchTreeDispatchGenerator :: DispatchTree → String → Continuable Maybe Pattern
function dispatchTreeDispatchGenerator(tree){return function(pathHead){return function(cont){var i,l,patterns,decoded
  assert(tree,"dispatchTreeDispatchGenerator: tree")
  if(pathHead==="")cont(methodBranch(tree))
  pathHead=pathHead.slice(0,-1) // lose the trailing slash
  patterns=tree[1]||tree.cases
  if(isBranchType(tree)){
    for(i=0,l=patterns.length;i<l;i++){
      if(patterns[i].test===pathHead
      || patterns[i].regex&&patterns[i].regex.source===pathHead
      || patterns[i].catchall&&pathHead==='(catchall)'){
        cont(pathBranch(pathHead,patterns[i]));return}}}
  cont() // decline
  function methodBranch(tree){
    assert(tree,"tree")
    return Pattern("",ChoicePoint(METHOD,
             [Pattern('GET',Handler(dispatchTreeGET(tree)))
             ,Pattern('POST',Handler(dispatchTreePOST(tree)))]))}
  function pathBranch(pathHead,pattern){
    return Pattern(pathHead+"/",ChoicePoint(PATHHEAD,
           [{generator:dispatchTreeDispatchGenerator(pattern.then)}]))}}}}

function dispatchTreeGET(tree){return function(req,resp){var body
  body=JSON.stringify(
    [typeNames[tree[0]]
    ,tree[0]==HANDLER?g(tree):f(tree[1])])+'\n'
  resp.writeHead(200,{'Content-Type':'text/json','Content-Length':body.length})
  resp.write(body)
  resp.end()
  function f(patterns){var ret=[]
    if(!(patterns instanceof Array)){
      ret.push('patterns array is not an array (this is likely a bug in a generator).')
      return ret}
    patterns.forEach(function(pattern){
      if(typeof pattern.test === 'string')
        ret.push(['match'+(pattern.generated?' (generated)':'')
                 ,pattern.test
                 ,pattern.test?encodeURIComponent(pattern.test)+'/'
                              :'(emptyPath)/'])
      else if(pattern.generator!==undefined || pattern.uri && pattern.error)
        ret.push(pattern.uri?['generator'+(pattern.error?' (load failed with "'+pattern.error+'")'
                                                        :pattern.generator?'':' (pending function load)'),pattern.uri,pattern.args]
                            :['generator (generated)'])
      else if(pattern.catchall)
        ret.push([null
                 ,'(catchall)/'])
      else if(pattern.regex)
        ret.push(['regex'
                 ,pattern.regex.source
                 ,encodeURIComponent(pattern.regex.source)+'/']) // XXX this can create URI collisions with literal patterns
      else
        ret.push(['unrepresentable pattern (this is either a bug here or in a generator): '+JSON.stringify(pattern)])})
    return ret}
  function g(handler_thing){
    if(!handler_thing[2]){ // no URI => must be generated
      return ["RequestHandler","(generated)"]}
    if(handler_thing[1]){ // it has a handler function
      return ["RequestHandler",handler_thing[1].name].concat(handler_thing.slice(2))}
    if(handler_thing[5] instanceof process.Promise){
      return ["RequestHandler","(pending)"].concat(handler_thing.slice(2))}
    // otherwise there's no handler function and no promise, so it must be an error
    return ["RequestHandler","error",handler_thing[5]]}}}

/* We have handlers, like dispatchTreeGET above, which are already very 
 * specific, containing the HTTP method name in the function.  There is 
 * no need to test the method in the handler since this is done in the 
 * dispatcher.  In general, pushing as much as possible out of the 
 * handler code and into the dispatch tree is a goal; this keeps things 
 * contained in one place and makes it possible to reason holistically 
 * about security and other server properties by examining the dispatch 
 * tree alone.
 */

function dispatchTreePUT(req,resp){
  // The dispatch tree is defined in a PUT request by a JSON document 
  // which describes a node of the tree.
  /* Unimplemented */
  throw new Error('unimplemented')
  }

/* loads a module from an absolute or relative (to cwd) path */
// loadAsync :: (path::String) → Continuable Either Error (exports::Object)
function loadAsync(path){return function(cont){
  dump(path)
  file.readFile(path,'utf-8')(E.replaceRight(cont)(_eval))
  function _eval(code){var exports={}
    function _require(x){var d,e
      d=+new Date
      e=require(x)
      dump('require: sync require '+x+' took '+(new Date-d)+'ms')
      return e}
    dump('read '+code.length+' bytes of code')
    code="(function __module_wrapper__(require,exports,nhttpd){"
        +code
        +"\n;return exports})"
    //dump('eval code: '+code)
    try{process.compile(code,path)(_require,exports,nhttpd)}
    catch(e){dump('caught an exception');dump(sys.inspect(e));cont(Left(e));return}
    dbg3('eval succeeded')
    cont(Right(exports))}}}

// modules: our require(), which pulls from a cache or is synchronous
// shared state: "shared x" in args
// filesystem access: "path x" in args
// anything that doesn't have shared state or a path in its arguments can be relocated transparently

// in the case of shared state or filesystem access, relocation is not possible
// how could we make something like revstore horizontally scalable?
// since there is a move command, and the history of a resource can span any number of paths, it is complex
// however only asof/ provides access to the historical revisions, so that might be the first horizontal split to make

// there should probably be a relocatable:false opt-out on generators and handlers, and otherwise the system should detect any fs reads or writes (i.e. by providing a custom fs module) and if they are not under a declared 'path' argument then it should print a warning or generate some kind of error.
// this will prevent accidentally writing code that can't be relocated.

// Idea: we could statically analyze a module and find all require() 
// calls which have constant arguments.  If there are any which do not, 
// those could either throw an error, load synchronously, or something 
// else.  In the constant argument cases, we resolve and cache all such 
// files before running the module.  Then when the module calls 
// require() we simply go to the cache.  This allows dynamic calls to 
// require (e.g. if(test){require('foo');...}) but does not allow calls 
// to require with arguments known only at runtime (e.g.  
// require('foo/'+some_user_input)).

functionFromURI.cache={}
// functionFromURI :: (String, String) → Continuable Either Error (_ → _)
function functionFromURI(path,fname){return function(cont){var f,a
  if(f=functionFromURI.cache[path+','+fname])return cont(Right(f))
  loadAsync(path+'.js')(E.replaceRight(cont)(
  function(exports){
    f=exports[fname]
    if(typeof f !== 'function'){
      cont(Left(fname+' is not a function (in '+path+')'));return}
    functionFromURI.cache[path+','+fname]=f
    cont(Right(f))}))}}

// branchFromJSON :: String → Either String Branch
function branchFromJSON(json){var obj
  try{obj=JSON.parse(json)}
  catch(e){return Left("Input could not be parsed as JSON. "+e.toString())}
  return branchFromObject(obj)}

function branchFromObject(obj){var tree,f,genobj,regex
  switch(obj[0]){
    case "match":
      if(typeof(obj[1])!=='string')return Left("Branch pattern must be a string (was "+obj[1]+")")
      tree=treeFromObject(obj[2])
      return applyRight(function(tree){return {test:obj[1],then:tree}})(tree)
    case "regex":
      if(typeof(obj[1])!=='string')return Left("Regex must be given as a string (was "+obj[1]+")")
      try{regex=new RegExp(obj[1])}catch(e){return Left(e)}
      //dump('regex.source: '+regex.source)
      return applyRight(function(tree){return {regex:regex,then:tree}})(treeFromObject(obj[2]))
    case "generator":
      genobj=obj[1]
      genobj.args=processArgs(genobj.args)
      genobj.pending=true
      functionFromURI(genobj.uri,genobj.fun)(either(
       function(e){
        genobj.error=e
        genobj.pending=false},
       function(f){
        genobj.generator=f.apply(null,genobj.args)
        genobj.pending=false}))
      return Right(genobj)
    case null: // a catchall branch, the null pattern matches anything
      return applyRight(function(tree){return {catchall:true,then:tree}})(treeFromObject(obj[1]))
    default:
      return Left('Unknown or unhandled branch type "'+obj[0]+'"')}}

function processArgs(args){
  return args.map(function(arg){
    if(typeof arg == 'string' || typeof arg == 'number')return arg
    if(arg[0]==='Shared')return sharedStateObject(arg[1])
    throw new Error("processArgs: unhandled argument type")})}

sharedStateObject.store={}
function sharedStateObject(name){
  if(Object.prototype.hasOwnProperty.call(sharedStateObject.store,name))return sharedStateObject.store[name]
  return sharedStateObject.store[name]={}}

function branchesFromArray(arr){var branches=[],i,l,branch
  if(!(arr instanceof Array))return Left("Dispatch branches must be provided as an array.")
  for(i=0,l=arr.length;i<l;i++){
    branch=branchFromObject(arr[i])
    if(E.isLeft(branch))return branch
    branches.push(branch[1])}
  return Right(branches)}

function handlerBranchFromObj(obj){var args,ret,callbacks
  if(!obj.fun||!obj.uri)return Left("Request handler URI or function name not specified.")
  obj.args=obj.args||[]
  args=processArgs(obj.args)
  callbacks=[]
  ret=[HANDLER,,obj.uri,obj.fun,obj.args,callbacks]
  if(obj.uri=='builtins')handlerBuiltins(obj)(either(handle_error,handle_handler))
  else functionFromURI(obj.uri,obj.fun)(either(handle_error,handle_f))
  return Right(ret)
  function handle_f(f){
    f.apply(null,args)(either(handle_error,handle_handler))}
  function handle_handler(handler){
    ret[1]=handler // the handler is now live for incoming requests
    fire_callbacks()
    ret.pop()}
  function handle_error(err){
    ret[6]=err
    dbg1('error loading request handler: '+err)
    fire_callbacks()}
  function fire_callbacks(){callbacks.forEach(function(callback){callback.call()});ret[5]=undefined}}

function handlerBuiltins(obj){return function(cont){var fs
  fs={handle404:handlers.handle404
     ,hello_world:handlers.hello_world
     }
  f=fs[obj.fun]
  if(!f)cont(Left("No such builtin handler: "+obj.fun))
  cont(Right(f))}}

function authBranch(specifier,cases){var branch,args,branches
  args=processArgs(specifier.args)
  branches=branchesFromArray(cases)
  if(!branches[0])return branches
  branches=branches[1]
  branch={0:AUTHN,pending:true,cases:branches,callbacks:[]}
  branch[1]=branches // XXX temporary
  functionFromURI(specifier.uri,specifier.fun)(either(handle_err,handle_f))
  return Right(branch)
  function handle_err(err){
    branch.error=err
    finish()}
  function handle_f(f){
    branch.auth_function=f.apply(null,args)
    finish()}
  function finish(){
    branch.pending=false
    branch.callbacks.forEach(function(callback){callback.call()})}}

function treeFromObject(arr){var n,branches,handler
  if(!(arr instanceof Array))return Left("Dispatch tree must be given as an array.")
  if((n=typeNames.indexOf(arr[0]))==-1)return Left("Unknown tree type "+arr[0])
  switch(n){
    case PATHHEAD:
    case FILENAME:
    case METHOD:
    case QUERY:
      return applyRight(function(branches){return [n,branches]})(branchesFromArray(arr[1]))
    case AUTHN:
      return authBranch(arr[1],arr[2])
    case HANDLER:
      return handlerBranchFromObj(arr[1]) // a handler is not really a branch at all; the terminology is quite muddy here already
    default:
      return Left("Unhandled tree type "+arr[0])}}

// example JSON document adding a branch to the tree:
// ["match"
// ,"styles/"
// ,["PathBranch",[["generator"
//                 ,"lib/generators/staticFileDispatch"
//                 ,"/var/www/static/javascript/"]]]]

// two branches are equivalent iff they have the same pattern type (e.g.  
// literal string, regex, catchall, etc) and if the pattern itself is 
// also a match (i.e. the regexes, literal string, etc are the same).
// Two generators are never equivalent, so removing generators must be 
// done (using DELETE) if desired, before adding the replacement.
// This behavior is the same as if old branches were never removed from 
// the tree, as the only ones which are removed are those that would be 
// shadowed by the newly inserted branch anyway.
function branchEq(a,b){
  if(a.test!==b.test
  || a.regex!==b.regex
  || a.catchall!==b.catchall)return false
  return true}

function insertOrReplaceBranch(branches,branch){var i,l
  for(i=0,l=branches.length;i<l;i++){
    if(branchEq(branches[i],branch)){
      branches.splice(i,1,branch);return}}
  branches.unshift(branch)}

function dispatchTreePOST(tree){return function dispatchTreePOST(req,resp){
  getBody(req,resp,function(body){var obj,branch
    branch=branchFromJSON(body)
    //dump(body)
    //dump(branch)
    if(!branch[0]&&!branch[1])throw new Error("")
    if(!branch[0]){dump(branch);handlers.error(400,"Bad Request",branch[1])(req,resp);return}
    insertOrReplaceBranch(tree[1],branch[1])
    // on success we send the response that a GET would get
    assert(tree,"tree")
    dispatchTreeGET(tree)(req,resp)})}}

// Each tree type has an implicit default:
// PATHHEAD => 404
// AUTHN => 401 Authorization Required
// METHOD => 405 method not allowed
// UA => ...
// REMOTE => 403 Forbidden

// These can be overridden by an explicit catchall.

treeFallthrough=[]
treeFallthrough[METHOD]=handlers.handle405
treeFallthrough[PATHHEAD]=handlers.handle404
treeFallthrough[UA]=handlers.unimplemented // not really sure what to do here
treeFallthrough[AUTHN]=handlers.handle401
treeFallthrough[COOKIE]=handlers.unimplemented
treeFallthrough[REMOTE]=handlers.handle403
treeFallthrough[FILENAME]=handlers.unimplemented
treeFallthrough[QUERY]=handlers.unimplemented // 400?

/* HTTP Server */

// Handling actual requests: we examine the request and match it 
// against the dispatch tree.  This always yields a ResponseHandler 
// either by an explicit match in the tree or by a fall-through.

server=http.createServer(function(req,resp){var qpos
  dump([req.connection.remoteAddress,rfc.ISO_date_time_space(new Date),req.method,req.url,req.headers['referer'],req.headers['user-agent']].join(' '))
  //dump([req.method+' '+req.url,req.headers]) // debugging
  // pathTail includes the path without the leading slash
  // queryString includes the query string, if any, without the question mark
  qpos=req.url.lastIndexOf('?')
  if(qpos>-1){req.pathTail=req.url.slice(1,qpos);req.queryString=req.url.slice(qpos+1)}
  else {req.queryString="";req.pathTail=req.url.slice(1)}
  dispatch(dispatchTree,req,resp)})

server.listen(port,domain)
sys.puts("Server listening on "+domain+":"+port)

function dispatch(tree,req,resp){
  var patterns
    , x
    , i,l
    , pathTail_orig,pathHead,pos,pathTail_rem
    , fallthroughHandler
    , pathLast
  descent:for(;;){
    dbg2(typeNames[tree[0]]||'unknown tree type: '+tree[0])
    switch(tree[0]){
      case HANDLER:
        if(tree[1]){
          dbg2('handler: '+tree[1].name)
          return tree[1](req,resp)}
        if(tree[6]){ // error loading the handler
          // handlers.error(500,"Internal Server Error","The request handler could not be loaded: "+tree[6])(req,resp) // information disclosure, maybe should depend on debug level
          handlers.error(500,"Internal Server Error","The request handler could not be loaded")(req,resp)
          dbg1('Dispatched to handler which failed with '+tree[6])
          return}
        if(tree[5]){ // pause any pending requests until the handler is ready
          // (this is probably not what we want to do in all cases... there should be a timeout at least)
          tree[5].push(function(){dispatch(tree,req,resp)})
          return}
        assert(false,'not reached')
      case METHOD:
        patterns=tree[1]
        dbg2('method: '+req.method)
        for(i=0,l=patterns.length;i<l;i++){
          dbg3(patterns[i])
          if(match(patterns[i],req.method)){
            dbg2('match')
            tree=patterns[i].then
            continue descent}}
        dbg2('no match: '+req.method)
        break
      case PATHHEAD:
        patterns=tree[1]
        pos=req.pathTail.indexOf('/') // find the next '/' if any
        fallthroughHandler=treeFallthrough[PATHHEAD]
        // set the pathHead and store the remaining pathTail on the request:
        //pathTail_orig=req.pathTail
        if(pos==-1){pathHead=req.pathTail;pathTail_rem=''}
        else{pathHead=req.pathTail.slice(0,pos+1);pathTail_rem=req.pathTail.slice(pos+1)}
        pathHead=decodeURIComponent(pathHead)
        dbg2('pathHead: "'+pathHead+'" pathTail_rem: "'+pathTail_rem+'"')
        for(i=0,l=patterns.length;i<l;i++){
          if(match(patterns[i],pathHead)){
            tree=patterns[i].then
            if(!patterns[i].catchall) req.pathTail=pathTail_rem // if this was a catchall, we didn't actually consume the path component
            continue descent}
          // if there is a generator, we call it
          // generator :: String → Continuable Maybe Pattern
          else if(patterns[i].pending){
            dbg2('pending generator')
            patterns[i].callbacks.push(function(){dispatch([PATHHEAD,patterns.slice(i)])})
            return}
          else if(patterns[i].generator){
            dbg2('generator')
            patterns[i].generator(pathHead)(makePatternHandler(patterns,[PATHHEAD,patterns.slice(i+1)],req,resp,function(){req.pathTail=pathTail_rem}))
            return}}
        break
      case AUTHN:
        // the user-provided function in tree[1] does the actual authentication, taking the request as its argument, and yields a string result, possibly asynchronously.
        // the string result of the authentication is what is then matched against.
        if(tree.pending){
          tree.callbacks.push(do_auth)
          return}
        do_auth()
        return
        function do_auth(){
          if(tree.error){
            handlers.error(500,"Server Error","Error loading authentication handler: "+tree.error)(req,resp)
            return}
          tree.auth_function(req,handle_auth_result)}
        function handle_auth_result(result){var i,l,patterns
          dbg2('auth result: '+result)
          patterns=tree.cases
          dbg3(patterns)
          for(i=0,l=patterns.length;i<l;i++){
            if(match(patterns[i],result)){
              dispatch(patterns[i].then,req,resp)
              return}}
          treeFallthrough[AUTHN](req,resp)}
      case UA:
        /* unimplemented */
      case COOKIE:
        /* unimplemented */
        return handlers.error(500,"Server Error","Unhandled dispatch tree branch type.")(req,resp)
      case REMOTE:
        patterns=tree[1]
        dbg2([req.connection.remoteAddress,patterns])
        for(i=0,l=patterns.length;i<l;i++){
          if(match(patterns[i],req.connection.remoteAddress)){
            tree=patterns[i].then
            continue descent}}
        break
      case FILENAME:
        patterns=tree[1]
        pathLast=req.url.slice(req.url.indexOf('/'))
        dbg2('pathLast: '+pathLast)
        for(i=0,l=patterns.length;i<l;i++){
          if(match(patterns[i],pathLast)){
            tree=patterns[i].then
            continue descent}}
        break
      case QUERY:
        patterns=tree[1]
        dbg2('queryString: '+req.queryString)
        for(i=0,l=patterns.length;i<l;i++){
          if(match(patterns[i],req.queryString)){
            tree=patterns[i].then
            continue descent}
          else if(patterns[i].generator){
            patterns[i].generator(req.queryString)(makePatternHandler(patterns,i,req,resp))
            return}}
        break
      case HOST:
        patterns=tree[1]
        dbg2('Host: '+req.headers.host)
        for(i=0,l=patterns.length;i<l;i++){
          if(match(patterns[i],req.headers.host)){
            tree=patterns[i].then
            continue descent}
          else if(patterns[i].generator){
            patterns[i].generator(req.headers.host)(makePatternHandler(patterns,i,req,resp))
            return}}
        break
      default:
        // This should never happen.
        throw new Error("Something is deeply wrong with the dispatch tree (should never happen)")}
    // if we get here, a 'break' was hit above, which means we fell 
    // through a list of patterns without hitting a match:
    if(!treeFallthrough[tree[0]]){
      throw new Error('Fatal error: no fallthrough for branch type: '+tree[0])}
    return treeFallthrough[tree[0]](req,resp)}}

function match(pattern,subject){
  return pattern.test===subject
      || pattern.regex&&pattern.regex.test(subject)
      || pattern.catchall}

function makePatternHandler(patterns,remainder,req,resp,extra){return function _patternHandler(pattern){
  if(!pattern){
    dbg2('generator declined')
    dispatch(remainder,req,resp)
    return}
  dbg2('dispatching through generated '+(pattern.transient?'transient pattern':'pattern with test='+pattern.test))
  // If the generator doesn't want the pattern used more than once, it can set the transient property to true
  if(!pattern.transient){
    pattern.generated=true
    pattern.destroy=makePatternDestroy(patterns,pattern)
    // the new pattern cannot mask any preceding literal match or generator, since it must have fallen through them to reach the generator that produced it, so it should be safe to put the new pattern at the top, speeding up future requests in the case where more-recently requested resources are more likely to be requested again
    patterns.unshift(pattern)}
  // note there is no test to see if the pattern actually matches the request
  if(extra){dbg2('calling extra function');extra()}
  dbg2('re-dispatching, pathTail: '+req.pathTail)
  dispatch(pattern.then,req,resp)}}

function makePatternDestroy(patterns,pattern){return function _patternDestroy(){var i,l
  for(i=0,l=patterns.length;i<l;i++) if(patterns[i]==pattern) patterns.splice(i,1)}}

function getBody(req,resp,f){var body=[]
  if(req.headers['expect'] && req.headers['expect'].indexOf('100-continue')!==-1){
    req.connection.setNoDelay()
    req.connection.write('HTTP/1.1 100 Continue\r\n\r\n')}
  req.setEncoding("utf8")
  req.addListener("data",function(chunk){
    body.push(chunk)})
  req.addListener("end",function(){
    body=body.join('')
    f(body)})}