// should be a little family of modules, one per supported spec, such as /rfc_util/{http,mime,...}
// or just one big module
// on the exports of each we have parsers and generators for all the interesting little string microsyntaxes
// the method names are based on the productions used in the relevant RFC
// there are also aliases where appopriate, e.g. in the HTML module we will add aliases by header names, each of which is a constructor taking some obvious arguments, and returning a string that can be used as that header.
// e.g. rfclib_http.headers.date :: Date → String
// or something like that, if it can be made convenient enough to be worth not having to look up the name of the production in the docs...

// from RFC 2616
// 3.3.1 Full Date
// 
//    HTTP applications have historically allowed three different formats
//    for the representation of date/time stamps:
// 
//       Sun, 06 Nov 1994 08:49:37 GMT  ; RFC 822, updated by RFC 1123
//       Sunday, 06-Nov-94 08:49:37 GMT ; RFC 850, obsoleted by RFC 1036
//       Sun Nov  6 08:49:37 1994       ; ANSI C's asctime() format
// 
//    The first format is preferred as an Internet standard and represents
//    a fixed-length subset of that defined by RFC 1123 [8] (an update to
//    RFC 822 [9]). The second format is in common use, but is based on the
//    obsolete RFC 850 [12] date format and lacks a four-digit year.
//    HTTP/1.1 clients and servers that parse the date value MUST accept
//    all three formats (for compatibility with HTTP/1.0), though they MUST
//    only generate the RFC 1123 format for representing HTTP-date values
//    in header fields. See section 19.3 for further information.
// 
//       Note: Recipients of date values are encouraged to be robust in
//       accepting date values that may have been sent by non-HTTP
//       applications, as is sometimes the case when retrieving or posting
//       messages via proxies/gateways to SMTP or NNTP.
// 
//    All HTTP date/time stamps MUST be represented in Greenwich Mean Time
//    (GMT), without exception. For the purposes of HTTP, GMT is exactly
//    equal to UTC (Coordinated Universal Time). This is indicated in the
//    first two formats by the inclusion of "GMT" as the three-letter
//    abbreviation for time zone, and MUST be assumed when reading the
//    asctime format. HTTP-date is case sensitive and MUST NOT include
//    additional LWS beyond that specifically included as SP in the
//    grammar.
// 
//        HTTP-date    = rfc1123-date | rfc850-date | asctime-date
//        rfc1123-date = wkday "," SP date1 SP time SP "GMT"
//        rfc850-date  = weekday "," SP date2 SP time SP "GMT"
//        asctime-date = wkday SP date3 SP time SP 4DIGIT
//        date1        = 2DIGIT SP month SP 4DIGIT
//                       ; day month year (e.g., 02 Jun 1982)
//        date2        = 2DIGIT "-" month "-" 2DIGIT
//                       ; day-month-year (e.g., 02-Jun-82)
//        date3        = month SP ( 2DIGIT | ( SP 1DIGIT ))
//                       ; month day (e.g., Jun  2)
//        time         = 2DIGIT ":" 2DIGIT ":" 2DIGIT
//                       ; 00:00:00 - 23:59:59
//        wkday        = "Mon" | "Tue" | "Wed"
//                     | "Thu" | "Fri" | "Sat" | "Sun"
//        weekday      = "Monday" | "Tuesday" | "Wednesday"
//                     | "Thursday" | "Friday" | "Saturday" | "Sunday"
//        month        = "Jan" | "Feb" | "Mar" | "Apr"
//                     | "May" | "Jun" | "Jul" | "Aug"
//                     | "Sep" | "Oct" | "Nov" | "Dec"
// 
//       Note: HTTP requirements for the date/time stamp format apply only
//       to their usage within the protocol stream. Clients and servers are
//       not required to use these formats for user presentation, request
//       logging, etc.

var wkdays=["Sun","Mon","Tue","Wed","Thu","Fri","Sat","Sun"] // Sunday is at [0] and at [7]
  , months=["Jan","Feb","Mar","Apr","May","Jun","Jul","Aug","Sep","Oct","Nov","Dec"]
function pad2(n){return n>9?''+n:'0'+n}
function rfc1123_date(d){return wkdays[d.getUTCDay()]+', '+pad2(d.getUTCDate())+' '+months[d.getUTCMonth()]+' '+d.getUTCFullYear()}

function ISO_date(d){return d.getUTCFullYear()+'-'+pad2(d.getUTCMonth()+1)+'-'+pad2(d.getUTCDate())}

function ISO_date_time_T    (d){return ISO_date(d)+'T'+pad2(d.getUTCHours())+':'+pad2(d.getUTCMinutes())+':'+pad2(d.getUTCSeconds())+'Z'}
function ISO_date_time_space(d){return ISO_date(d)+' '+pad2(d.getUTCHours())+':'+pad2(d.getUTCMinutes())+':'+pad2(d.getUTCSeconds())+'Z'}

exports.rfc1123_date=rfc1123_date
exports.ISO_date=ISO_date
exports.ISO_date_time_T=ISO_date_time_T
exports.ISO_date_time_space=ISO_date_time_space


// From RFC 2045
// 6.8.  Base64 Content-Transfer-Encoding
// 
//    The Base64 Content-Transfer-Encoding is designed to represent
//    arbitrary sequences of octets in a form that need not be humanly
//    readable.  The encoding and decoding algorithms are simple, but the
//    encoded data are consistently only about 33 percent larger than the
//    unencoded data.  This encoding is virtually identical to the one used
//    in Privacy Enhanced Mail (PEM) applications, as defined in RFC 1421.
// 
//    A 65-character subset of US-ASCII is used, enabling 6 bits to be
//    represented per printable character. (The extra 65th character, "=",
//    is used to signify a special processing function.)
// 
//    NOTE:  This subset has the important property that it is represented
//    identically in all versions of ISO 646, including US-ASCII, and all
//    characters in the subset are also represented identically in all
//    versions of EBCDIC. Other popular encodings, such as the encoding
//    used by the uuencode utility, Macintosh binhex 4.0 [RFC-1741], and
//    the base85 encoding specified as part of Level 2 PostScript, do not
//    share these properties, and thus do not fulfill the portability
//    requirements a binary transport encoding for mail must meet.
// 
//    The encoding process represents 24-bit groups of input bits as output
//    strings of 4 encoded characters.  Proceeding from left to right, a
//    24-bit input group is formed by concatenating 3 8bit input groups.
//    These 24 bits are then treated as 4 concatenated 6-bit groups, each
//    of which is translated into a single digit in the base64 alphabet.
//    When encoding a bit stream via the base64 encoding, the bit stream
//    must be presumed to be ordered with the most-significant-bit first.
//    That is, the first bit in the stream will be the high-order bit in
//    the first 8bit byte, and the eighth bit will be the low-order bit in
//    the first 8bit byte, and so on.
// 
//    Each 6-bit group is used as an index into an array of 64 printable
//    characters.  The character referenced by the index is placed in the
//    output string.  These characters, identified in Table 1, below, are
//    selected so as to be universally representable, and the set excludes
//    characters with particular significance to SMTP (e.g., ".", CR, LF)
//    and to the multipart boundary delimiters defined in RFC 2046 (e.g.,
//    "-").
// 
// 
//                     Table 1: The Base64 Alphabet
// 
//      Value Encoding  Value Encoding  Value Encoding  Value Encoding
//          0 A            17 R            34 i            51 z
//          1 B            18 S            35 j            52 0
//          2 C            19 T            36 k            53 1
//          3 D            20 U            37 l            54 2
//          4 E            21 V            38 m            55 3
//          5 F            22 W            39 n            56 4
//          6 G            23 X            40 o            57 5
//          7 H            24 Y            41 p            58 6
//          8 I            25 Z            42 q            59 7
//          9 J            26 a            43 r            60 8
//         10 K            27 b            44 s            61 9
//         11 L            28 c            45 t            62 +
//         12 M            29 d            46 u            63 /
//         13 N            30 e            47 v
//         14 O            31 f            48 w         (pad) =
//         15 P            32 g            49 x
//         16 Q            33 h            50 y
// 
//    The encoded output stream must be represented in lines of no more
//    than 76 characters each.  All line breaks or other characters not
//    found in Table 1 must be ignored by decoding software.  In base64
//    data, characters other than those in Table 1, line breaks, and other
//    white space probably indicate a transmission error, about which a
//    warning message or even a message rejection might be appropriate
//    under some circumstances.
// 
//    Special processing is performed if fewer than 24 bits are available
//    at the end of the data being encoded.  A full encoding quantum is
//    always completed at the end of a body.  When fewer than 24 input bits
//    are available in an input group, zero bits are added (on the right)
//    to form an integral number of 6-bit groups.  Padding at the end of
//    the data is performed using the "=" character.  Since all base64
//    input is an integral number of octets, only the following cases can
//    arise: (1) the final quantum of encoding input is an integral
//    multiple of 24 bits; here, the final unit of encoded output will be
//    an integral multiple of 4 characters with no "=" padding, (2) the
//    final quantum of encoding input is exactly 8 bits; here, the final
//    unit of encoded output will be two characters followed by two "="
//    padding characters, or (3) the final quantum of encoding input is
//    exactly 16 bits; here, the final unit of encoded output will be three
//    characters followed by one "=" padding character.
// 
//    Because it is used only for padding at the end of the data, the
//    occurrence of any "=" characters may be taken as evidence that the
//    end of the data has been reached (without truncation in transit).  No
//    such assurance is possible, however, when the number of octets
//    transmitted was a multiple of three and no "=" characters are
//    present.
// 
//    Any characters outside of the base64 alphabet are to be ignored in
//    base64-encoded data.
// 
//    Care must be taken to use the proper octets for line breaks if base64
//    encoding is applied directly to text material that has not been
//    converted to canonical form.  In particular, text line breaks must be
//    converted into CRLF sequences prior to base64 encoding.  The
//    important thing to note is that this may be done directly by the
//    encoder rather than in a prior canonicalization step in some
//    implementations.
// 
//    NOTE: There is no need to worry about quoting potential boundary
//    delimiters within base64-encoded bodies within multipart entities
//    because no hyphen characters are used in the base64 encoding.


var base64_alphabet="ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/"

// version for 8-bit input (i.e. only the low 8 bits is used in each 16-bit code unit in a JavaScript string)
// doesn't do the 76-character line length limit
function rfc2045_base64_encode(input){var quantum,pos,l,output
 output=""
 // handling 24 bits at a time, up to last complete "quantum" of 3 octets
 for(pos=0,l=input.length-2;pos<l;pos+=3){
  // "a 24-bit input group is formed by concatenating 3 8bit input groups"
  quantum=input.charCodeAt( pos )<<16
         |input.charCodeAt(pos+1)<<8
         |input.charCodeAt(pos+2)
  // "These 24 bits are then treated as 4 concatenated 6-bit groups, each
  //  of which is translated into a single digit in the base64 alphabet."
  output+=base64_alphabet[quantum>>18]
         +base64_alphabet[quantum>>12&0x3F]
         +base64_alphabet[quantum>>6 &0x3F]
         +base64_alphabet[quantum    &0x3F]}
 // handling incomplete quantum of 1 or 2 extra input bytes
 if(input.length-pos==1){
  output+=base64_alphabet[input.charCodeAt(pos)>>2]        // first 6 bits of the last byte
         +base64_alphabet[input.charCodeAt(pos)<<4&0x30]   // last two bits of data plus four bits of zeros
         +"=="}                                            // padding
 else if(input.length-pos==2){
  output+=base64_alphabet[input.charCodeAt(pos)>>2]        // first 6 bits
         +base64_alphabet[(input.charCodeAt(pos)<<4&0x3F)  // last two bits of penultimate input octet
                          |input.charCodeAt(pos+1)>>4]     // with top four of final octet
         +base64_alphabet[input.charCodeAt(pos+1)<<2&0x3C] // last four bits
         +"="}                                             // padding
 return output}

exports.rfc2045_base64_encode=rfc2045_base64_encode
exports.base64_encode=rfc2045_base64_encode
